@@ -5,6 +5,9 @@ use anyhow::{bail, Error};
 use fxhash::FxHashMap;
 use fxhash::FxHashSet;
 
+#[cfg(test)]
+mod tests;
+
 #[derive(Debug, Default)]
 struct PlanBuilder {
     /// `(src, dst)`
@@ -42,7 +45,7 @@ where
                 None => {}
             }
 
-            self.add_to_graph(&mut builder, module.id, &mut vec![]);
+            self.add_to_graph(&mut builder, module.id, &mut vec![module.id]);
         }
 
         Ok((
@@ -61,6 +64,9 @@ where
         module_id: ModuleId,
         path: &mut Vec<ModuleId>,
     ) {
+        if cfg!(test) {
+            log::debug!("Adding {:?} to the graph (path = {:?})", module_id, path);
+        }
         let visited = builder.all.contains(&module_id);
         let cycle_rpos = if visited {
             path.iter().rposition(|v| *v == module_id)
@@ -73,6 +79,23 @@ where
             builder.cycles.push(cycle);
         }
 
+        let prev_last = *path.last().unwrap();
+        // Prevent infinite recursion.
+        if !builder.tracked.insert((prev_last, module_id)) {
+            // This is a hack
+            //
+            // TODO(kdy1): Use proper logic for `builder.tracked` and remove this hack.
+            if let Some(cycle) = builder
+                .cycles
+                .iter_mut()
+                .find(|cycle| cycle.contains(&prev_last))
+            {
+                cycle.push(module_id);
+            }
+
+            return;
+        }
+
         path.push(module_id);
 
         if !visited {
@@ -95,21 +118,7 @@ where
 
             builder.graph.add_edge(module_id, src.module_id, ());
 
-            // Prevent infinite loops.
-            if builder.tracked.insert((module_id, src.module_id)) {
-                self.add_to_graph(builder, src.module_id, path);
-            } else {
-                // This is a hack
-                //
-                // TODO(kdy1): Use proper logic for `builder.tracked` and remove this hack.
-                if let Some(cycle) = builder
-                    .cycles
-                    .iter_mut()
-                    .find(|cycle| cycle.contains(&module_id))
-                {
-                    cycle.push(src.module_id);
-                }
-            }
+            self.add_to_graph(builder, src.module_id, path);
         }
 
         let res = path.pop();
